import argparse
import os
import re
from pathlib import Path

from .colors import decolored
from .envs import shell_cmd

# Constants
OUTPUT_FILE = "cli.log"
LAST_LINES = 1000
MAX_LINES = 10000

TMUX = "TMUX"
CWD = Path.cwd()

# "$", "(\w+)", "~", "/", "<user>@<host>"
RE_MARK = r"[\$]"  # starts with $
RE_PATH = r"[\~\/]"  # starts with ~ or /
RE_CENV = r"\(\w+\)"  # starts with (env)
RE_USER = r"\w+\@\w+"  # starts with user@host
CMD_PATTERNS = [rf"^\s*({RE_MARK}|{RE_PATH}|{RE_CENV}|{RE_USER})"]
TMUX_CMD_PATTERNS = [" -m tclogger.tmux"]


class CmdPromptChecker:
    """Check command prompt line based on regex."""

    def __init__(self, pattern: str = None) -> None:
        if pattern:
            self.cmd_re = re.compile(pattern)
        else:
            self.cmd_re = re.compile("|".join(p for p in CMD_PATTERNS))

    def is_cmd_prompt(self, line: str) -> bool:
        return bool(self.cmd_re.match(line))

    def is_tmux_cmd(self, cmd: str) -> bool:
        return any(p in cmd for p in TMUX_CMD_PATTERNS)


class TmuxLogger:
    """Capture and log tmux pane commands history."""

    def __init__(
        self,
        output_file: str = OUTPUT_FILE,
        cmd_num: int = None,
        last_lines: int = LAST_LINES,
        max_lines: int = MAX_LINES,
        cmd_only: bool = False,
        must_have_output: bool = True,
        cmd_pattern: str = None,
        include_pattern: str = None,
        exclude_pattern: str = None,
        skip_tmux_cmds: bool = True,
    ) -> None:
        """
        Args:
        - `output_file`: Output file name (default: cli.log)
        - `cmd_num`: Number of recent cmds to include (default: None, will use all cmds from last_lines)
        - `last_lines`: Number of history lines to capture from tmux (default: 1000)
        - `max_lines`: Maximum number of lines to capture when searching for cmds (default: 10000)
        - `cmd_only`: If True, only capture cmds without outputs
        - `must_have_output`: If True, only keep cmd/output pairs where output has non-whitespace content
        - `cmd_pattern`: Custom regex pattern to identify cmd lines (default: None, uses built-in pattern)
        - `include_pattern`: Regex pattern to filter cmds to include (default: None, includes all)
        - `exclude_pattern`: Regex pattern to filter cmds to exclude (default: None, excludes none)
        - `skip_tmux_cmds`: If True, exclude tmux commands and outputs (default: True)
        """
        self.output_file = output_file
        self.cmd_num = cmd_num
        self.last_lines = last_lines
        self.max_lines = max_lines
        self.cmd_only = cmd_only
        self.must_have_output = must_have_output
        self.cmd_pattern = cmd_pattern
        self.include_pattern = include_pattern
        self.exclude_pattern = exclude_pattern
        self.skip_tmux_cmds = skip_tmux_cmds
        self.current_lines = last_lines
        self.log_path = CWD / self.output_file
        self.cmd_checker = CmdPromptChecker(pattern=cmd_pattern)

    @staticmethod
    def _has_effective_output(output: str) -> bool:
        return bool(output and output.strip())

    @staticmethod
    def is_in_tmux() -> bool:
        return TMUX in os.environ

    def capture_pane_to_file(self) -> bool:
        """Capture tmux pane output to file via `tmux capture-pane`.

        Returns:
            True if successful, False otherwise
        """
        if not self.is_in_tmux():
            return False

        try:
            cmd = f"tmux capture-pane -p -J -S -{self.current_lines} > {self.log_path}"
            # -p    : print to stdout
            # -S -N : N lines starting from bottom
            # -J    : join wrapped lines
            shell_cmd(cmd, showcmd=False)
            return True
        except Exception as e:
            return False

    def parse_cmds_outputs_from_file(self) -> list[tuple[str, str]]:
        """Parse cmds from tmux output file by search cmd prompts.

        Returns: list of (cmd, output) tuples
        """
        try:
            with open(self.log_path, "r", encoding="utf-8", errors="ignore") as f:
                output = f.read()
        except Exception:
            return []

        if not output:
            return []

        lines = output.split("\n")
        results = []

        current_cmd = None
        current_output = []

        for line in lines:
            line_rstrip = line.rstrip()
            if self.cmd_checker.is_cmd_prompt(line_rstrip):
                # Save previous cmd and output
                if current_cmd is not None:
                    results.append((current_cmd.rstrip(), "\n".join(current_output)))
                # Start new cmd (use whole line as cmd)
                current_cmd = line_rstrip
                current_output = []
            else:
                # This is output from current cmd
                if current_cmd is not None:
                    current_output.append(line_rstrip)

        # Append last cmd
        if current_cmd is not None:
            results.append((current_cmd.rstrip(), "\n".join(current_output)))

        # Remove tmux cmds
        if self.skip_tmux_cmds and results:
            results = [
                (cmd, output)
                for cmd, output in results
                if not self.cmd_checker.is_tmux_cmd(cmd)
            ]

        # Keep only pairs that have effective output
        if self.must_have_output and results:
            results = [
                (cmd, output)
                for cmd, output in results
                if self._has_effective_output(output)
            ]

        # Apply include pattern filter
        if self.include_pattern:
            include_re = re.compile(self.include_pattern)
            results = [
                (cmd, output) for cmd, output in results if include_re.search(cmd)
            ]

        # Apply exclude pattern filter
        if self.exclude_pattern:
            exclude_re = re.compile(self.exclude_pattern)
            results = [
                (cmd, output) for cmd, output in results if not exclude_re.search(cmd)
            ]

        # Return last N cmds
        if self.cmd_num is None:
            return results
        else:
            return results[-self.cmd_num :] if results else []

    def log(self) -> None:
        """Create log file with recent cmd history."""
        cmd_output_pairs = []

        # capture enough history to get the required number of cmds
        while self.current_lines <= self.max_lines:
            if not self.capture_pane_to_file():
                return
            cmd_output_pairs = self.parse_cmds_outputs_from_file()
            if self.cmd_num is None:
                break
            if len(cmd_output_pairs) >= self.cmd_num:
                break
            if self.current_lines < self.max_lines:
                self.current_lines = min(self.current_lines * 2, self.max_lines)
            else:
                break

        if not cmd_output_pairs:
            return

        # concat and decolor
        log_lines = []
        for cmd, output in cmd_output_pairs:
            log_lines.append(f"{cmd}\n{output}\n")
        log_content = "".join(log_lines)
        log_content = decolored(log_content)

        # save to file
        with open(self.log_path, "w", encoding="utf-8") as f:
            f.write(log_content)


class TmuxLoggerArgParser:
    def __init__(self) -> None:
        self.parser = argparse.ArgumentParser(
            description="Create log file with tmux cmd and outputs history"
        )
        self._add_arguments()

    def _add_arguments(self) -> None:
        self.parser.add_argument(
            "-o",
            "--output",
            type=str,
            default=OUTPUT_FILE,
            help=f"Output file (default: {OUTPUT_FILE})",
        )
        self.parser.add_argument(
            "-n",
            "--cmd-num",
            type=int,
            default=None,
            help="Number of recent cmds to include (default: None, will use all cmds from last-lines)",
        )
        self.parser.add_argument(
            "-l",
            "--last-lines",
            type=int,
            default=LAST_LINES,
            help=f"Number of history lines to capture from tmux (default: {LAST_LINES})",
        )
        self.parser.add_argument(
            "-m",
            "--max-lines",
            type=int,
            default=MAX_LINES,
            help=f"Maximum number of lines to capture when searching for cmds (default: {MAX_LINES})",
        )
        self.parser.add_argument(
            "-c",
            "--cmd-only",
            action="store_true",
            help="Capture cmds only, no outputs",
        )
        self.parser.add_argument(
            "-u",
            "--must-have-output",
            action=argparse.BooleanOptionalAction,
            default=True,
            help="Keep only cmd/output pairs whose output is non-whitespace (default: True)",
        )
        self.parser.add_argument(
            "-p",
            "--cmd-pattern",
            type=str,
            default=None,
            help="Custom regex pattern to identify cmd lines (default: None, uses built-in pattern)",
        )
        self.parser.add_argument(
            "-i",
            "--include-pattern",
            type=str,
            default=None,
            help="Regex pattern to filter cmds to include (default: None, includes all)",
        )
        self.parser.add_argument(
            "-e",
            "--exclude-pattern",
            type=str,
            default=None,
            help="Regex pattern to filter cmds to exclude (default: None, excludes none)",
        )

    def parse_args(self) -> argparse.Namespace:
        return self.parser.parse_args()


def log_tmux() -> None:
    """Main entry point for the tmux CLI tool."""
    arg_parser = TmuxLoggerArgParser()
    args = arg_parser.parse_args()

    tmlogger = TmuxLogger(
        output_file=args.output,
        cmd_num=args.cmd_num,
        last_lines=args.last_lines,
        max_lines=args.max_lines,
        cmd_only=args.cmd_only,
        must_have_output=args.must_have_output,
        cmd_pattern=args.cmd_pattern,
        include_pattern=args.include_pattern,
        exclude_pattern=args.exclude_pattern,
    )
    tmlogger.log()


if __name__ == "__main__":
    log_tmux()

    # python -m tclogger.tmux -n 5
    # python -m tclogger.tmux -n 5 -i "python"



from .types import StrsType, LIST_TYPES
from .maths import chars_len, is_str_float
from .fills import add_fills
from .colors import COLOR_TYPE, colored, decolored
from .logs import logclr

from copy import deepcopy
from typing import Literal

HAT_COLOR = logclr.NOTE
HEADER_COLOR = logclr.MESG
CELL_COLOR = logclr.FILE
SEPR_COLOR = logclr.DBUG
BOUND_COLOR = logclr.DBUG
VERT_COLOR = logclr.DBUG
TOTAL_COLOR = logclr.OKAY

SEPR = "-"
CSEPR = colored(SEPR, SEPR_COLOR)
VERT = "|"
CVERT = colored(VERT, VERT_COLOR)
WERT = f" | "


def is_listable(val) -> bool:
    return isinstance(val, LIST_TYPES)


def norm_any_to_str_list(val) -> list[str]:
    if is_listable(val):
        return [str(v) for v in val]
    else:
        return [str(val)]


def norm_any_to_type_list(val) -> list[str]:
    if is_listable(val):
        return [type(v).__name__ for v in val]
    else:
        return [type(val).__name__]


def add_bounds(
    line: str,
    bound_char: str = VERT,
    is_colored: bool = False,
    bound_color: COLOR_TYPE = BOUND_COLOR,
) -> str:
    end_bound_char = bound_char[::-1]
    if not is_colored:
        return f"{bound_char} {line} {end_bound_char}"
    else:
        beg_bound_str = colored(bound_char, bound_color)
        end_bound_str = colored(end_bound_char, bound_color)
        return f"{beg_bound_str} {line} {end_bound_str}"


def align_to_fill_side(align: str) -> Literal["left", "right", "both"]:
    if align[0].lower() == "l":
        return "right"
    elif align[0].lower() == "r":
        return "left"
    else:
        return "both"


def determine_aligns_from_row(
    row: list, default_align: Literal["left", "right"] = "left"
) -> list[str]:
    aligns: list[str] = []
    for cell in row:
        if is_str_float(decolored(cell)):
            aligns.append("right")
        else:
            aligns.append(default_align)
    return aligns


def rows_to_table_str(
    rows: list[list],
    headers: list[str] = None,
    aligns: StrsType = None,
    default_align: Literal["left", "right"] = "left",
    sum_at_tail: bool = False,
    header_case: Literal["raw", "lower", "upper", "capitalize"] = "upper",
    header_wsch: Literal[" ", "_", "-", "", None] = "_",
    col_gap_len: int = 2,
    is_bounded: bool = False,
    bound_char: str = VERT,
    is_hatted: bool = True,
    hat_char: str = "=",
    is_colored: bool = True,
    hat_color: COLOR_TYPE = HAT_COLOR,
    header_color: COLOR_TYPE = HEADER_COLOR,
    cell_color: COLOR_TYPE = CELL_COLOR,
    sepr_color: COLOR_TYPE = SEPR_COLOR,
    bound_color: COLOR_TYPE = BOUND_COLOR,
) -> str:
    if not rows:
        return ""

    # set table headers and table rows
    if not headers:
        headers = norm_any_to_str_list(rows[0])
        table_rows = rows[1:]
    else:
        table_rows = rows[:]

    table_headers: list[str] = deepcopy(headers)

    if header_wsch is not None:
        table_headers = [h.replace(" ", header_wsch) for h in table_headers]

    # set header case
    hc = header_case.lower() if header_case else ""
    if hc.startswith("l"):
        table_headers = [h.lower() for h in table_headers]
    elif hc.startswith("u"):
        table_headers = [h.upper() for h in table_headers]
    elif hc.startswith("c"):
        table_headers = [h.capitalize() for h in table_headers]
    # else: raw

    cols = len(table_headers)

    # determine aligns
    if not aligns:
        aligns = determine_aligns_from_row(table_rows[-1], default_align=default_align)
    if len(aligns) < cols:
        aligns += [default_align] * (cols - len(aligns))

    # add total row
    is_sum_at_tail = sum_at_tail and cols > 1
    if is_sum_at_tail:
        sum_row: list[str] = []
        for i in range(cols):
            is_sum_valid = True
            col_sum = 0
            for row in table_rows:
                cell = row[i]
                if not cell or cell == "-":
                    continue
                elif isinstance(cell, (int, float)):
                    col_sum += cell
                elif cell.isdigit():
                    col_sum += int(cell)
                elif is_str_float(cell):
                    col_sum += float(cell)
                else:
                    is_sum_valid = False
                    break
            if is_sum_valid and col_sum != 0:
                sum_row.append(str(col_sum))
            else:
                sum_row.append("")
        if any(v for v in sum_row):
            is_sum_at_tail = True
            # prepend "Total" label to numeric columns
            for i in range(1, cols):
                if sum_row[i]:
                    sum_row[i - 1] = "Total"
                    break
            table_rows.append(sum_row)
        else:
            is_sum_at_tail = False

    # calc column widths
    table_headers_rows = [table_headers] + table_rows
    col_widths = [
        max(
            chars_len(decolored(row[i])) if i < len(row) else 0
            for row in table_headers_rows
        )
        for i in range(cols)
    ]

    # set colors
    sepr = SEPR
    if is_colored:
        table_headers = [colored(h, header_color) for h in table_headers]
        if is_sum_at_tail:
            colored_sum_row = [
                colored(cell, TOTAL_COLOR) if cell else cell for cell in table_rows[-1]
            ]
            colored_non_sum_rows = [
                [colored(cell, cell_color) for cell in row] for row in table_rows[:-1]
            ]
            table_rows = colored_non_sum_rows + [colored_sum_row]
        else:
            table_rows = [
                [colored(cell, cell_color) for cell in row] for row in table_rows
            ]
        sepr = colored(SEPR, sepr_color)

    # set sep lines and col chars
    sep_lines = [sepr * col_widths[i] for i in range(cols)]

    jcol = " " * col_gap_len
    col_join = jcol.join
    sep_join = (sepr * len(jcol)).join
    header_line_str = col_join(
        add_fills(
            text=table_headers[i],
            filler=" ",
            fill_side=align_to_fill_side(aligns[i]),
            is_text_colored=is_colored,
            total_width=col_widths[i],
        )
        for i in range(cols)
    )
    sep_line_str = sep_join(sep_lines)

    # add vertical bounds
    if is_bounded:
        header_line_str = add_bounds(
            header_line_str,
            bound_char=bound_char,
            is_colored=is_colored,
            bound_color=bound_color,
        )
        sep_line_str = add_bounds(
            sep_line_str,
            bound_char=bound_char,
            is_colored=is_colored,
            bound_color=bound_color,
        )

    # build row lines
    rows_lines = []
    for row in table_rows:
        row_line = col_join(
            add_fills(
                text=row[i],
                filler=" ",
                fill_side=align_to_fill_side(aligns[i]),
                is_text_colored=is_colored,
                total_width=col_widths[i],
            )
            for i in range(cols)
        )
        if is_bounded:
            row_line = add_bounds(
                row_line, bound_char=bound_char, is_colored=is_colored
            )
        rows_lines.append(row_line)

    # build final table string
    if is_sum_at_tail:
        row_lines_str = (
            "\n".join(rows_lines[:-1]) + f"\n{sep_line_str}\n{rows_lines[-1]}"
        )
    else:
        row_lines_str = "\n".join(rows_lines)

    table_str = f"{header_line_str}\n{sep_line_str}\n{row_lines_str}"

    # add hat
    if is_hatted:
        hat_len = chars_len(decolored(table_str.splitlines()[0]))
        hat_str = add_fills(filler=hat_char, total_width=hat_len)
        if is_colored:
            hat_str = colored(hat_str, hat_color)
        table_str = f"{hat_str}\n{table_str}\n{hat_str}"

    return table_str


def dict_to_rows(d: dict) -> list[list]:
    rows: list[list] = []
    for key, val in d.items():
        key_strs = norm_any_to_str_list(key)
        val_strs = norm_any_to_str_list(val)
        row = key_strs + val_strs
        rows.append(row)
    return rows


def dict_to_table_str(
    d: dict,
    key_headers: StrsType = None,
    val_headers: StrsType = None,
    aligns: StrsType = None,
    default_align: Literal["left", "right"] = "left",
    sum_at_tail: bool = False,
    header_case: Literal["raw", "lower", "upper", "capitalize"] = "upper",
    header_wsch: Literal[" ", "_", "-", "", None] = "_",
    col_gap_len: int = 2,
    is_bounded: bool = False,
    bound_char: str = VERT,
    is_hatted: bool = True,
    hat_char: str = "=",
    is_colored: bool = True,
    hat_color: COLOR_TYPE = HAT_COLOR,
    header_color: COLOR_TYPE = HEADER_COLOR,
    cell_color: COLOR_TYPE = CELL_COLOR,
    sepr_color: COLOR_TYPE = SEPR_COLOR,
    bound_color: COLOR_TYPE = BOUND_COLOR,
) -> str:
    if not d:
        return ""
    if not key_headers or not val_headers:
        k1, v1 = next(iter(d.items()))
        if not key_headers:
            key_headers = norm_any_to_str_list(k1)
        if not val_headers:
            val_headers = norm_any_to_str_list(v1)
    table_headers: list[str] = key_headers + val_headers
    table_rows = dict_to_rows(d)

    return rows_to_table_str(
        rows=table_rows,
        headers=table_headers,
        aligns=aligns,
        default_align=default_align,
        sum_at_tail=sum_at_tail,
        header_case=header_case,
        header_wsch=header_wsch,
        col_gap_len=col_gap_len,
        is_bounded=is_bounded,
        bound_char=bound_char,
        is_hatted=is_hatted,
        hat_char=hat_char,
        is_colored=is_colored,
        hat_color=hat_color,
        header_color=header_color,
        cell_color=cell_color,
        sepr_color=sepr_color,
        bound_color=bound_color,
    )
